/****************************************************************************
** 	Blowfish.cpp
**
**	Blowfish is a 64-bit symmetric block cipher with a variable length key,
**	up to 448-bit or 56 bytes, designed by Bruce Schneier.
**
**	BlowfishTranslator's blowfish implementation is based off of Raif S.  
**	Naffah's Java port of blowfish. 	
**
**	Copyright (C) 2000, Jason Parks <atreyu@acm.org>.  All rights reserved.
**
*****************************************************************************/
/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
 
#include <ByteOrder.h>
#include <Locker.h>
#include <Message.h>
#include <MenuField.h>
#include <MenuItem.h>
#include <PopUpMenu.h>
#include <TranslatorAddOn.h>
#include <TranslationKit.h>
#include <View.h>

#include <ctype.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "KeyConstants.h"

//constants
const int32 MIN_NOF_ROUNDS 		= 16;		// this is also the default value
const int32 MAX_NOF_ROUNDS 		= 20;
const int32 BLOCK_SIZE 			= 8;		// Blowfish block size in bytes
const int32 MAX_USER_KEY_LENGTH = 448 / 8;	// given in bytes from a value in bits
const int32 ROUNDS 				= 'rnds';

const uint32 BLOWFISH_ENCRYPT 	= 'blow';
const uint32 BLOWFISH_DECRYPT 	= 'fish';

//global variables
char translatorName[] = "Blowfish Translator";
char translatorInfo[] = "Encrypts/Decrypts files using blowfish encryption";
int32 translatorVersion = 100; 

int32 P[18];
int32 sKey[4 * 256];
int32 rounds = MIN_NOF_ROUNDS;

BLocker settings_lock("Blowfish settings lock");

//foward declarations
void Encrypt(BPositionIO *inSource, BPositionIO *outSource, const char *key);
void Decrypt(BPositionIO *inSource, BPositionIO *outSource, const char *key);
void MakeKeys(const char *key);
void BF_encrypt(int32 L, int32 R, int32 *out, int32 outOff);

//config view
class BlowfishView : public BView
{
	public:
		BlowfishView(BRect frame, const char *name, uint32 resize, uint32 flags)
		:BView(frame, name, resize, flags)
		{
			SetViewColor(220,220,220,0);
			SetLowColor(216, 216, 216);
			
//			mMenu = new BPopUpMenu("Rounds");
//	
//			BMenuItem *item;
//			mMenu->AddItem(item = new BMenuItem("16", new BMessage(ROUNDS + 16)));
//			item->SetMarked(true);
//			mMenu->AddItem(new BMenuItem("17", new BMessage(ROUNDS + 17)));
//			mMenu->AddItem(new BMenuItem("18", new BMessage(ROUNDS + 18)));
//			mMenu->AddItem(new BMenuItem("19", new BMessage(ROUNDS + 19)));
//			mMenu->AddItem(new BMenuItem("20", new BMessage(ROUNDS + 20)));
//	
//			mField = new BMenuField(BRect(20,70,110,90), "Rounds Field", "Rounds:", mMenu);
//			mField->SetViewColor(ViewColor());
//				
//			AddChild(mField);
		}
		
		~BlowfishView()
		{
		}
					
		virtual	void Draw(BRect rect)
		{
			SetFont(be_bold_font);
			DrawString("Blowfish Translator", BPoint(12, 25));
		}
		
		virtual	void MessageReceived(BMessage *msg)
		{
			switch(msg->what)
			{
				case (ROUNDS + 16):
				case (ROUNDS + 17):
				case (ROUNDS + 18):
				case (ROUNDS + 19):
				case (ROUNDS + 20):
				{
					settings_lock.Lock();
					//rounds = msg->what - ROUNDS;
					settings_lock.Unlock();
					break;
				}
				default:
				{
					BView::MessageReceived(msg);
					break;
				}
			}
		}
			
	private:
		BPopUpMenu * mMenu;
		BMenuField * mField;
};

//global functions
status_t MakeConfig(BMessage *ioExtension, BView **outView, BRect *outExtent)
{
	BlowfishView *bfv = new BlowfishView(BRect(0,0,200,100), "BLOWFISHTranslator Settings", B_FOLLOW_ALL, B_WILL_DRAW);
	*outView = bfv;
	*outExtent = bfv->Bounds();
	
	return B_OK;
}					

status_t Identify(BPositionIO * inSource,
				  const translation_format *inFormat,
				  BMessage * ioExtension,
				  translator_info *outInfo,
				  uint32 outType)
{
	status_t status = B_NO_TRANSLATOR;	
		
	if(outType == BLOWFISH_ENCRYPT)
	{
		outInfo->group = B_TRANSLATOR_NONE;
		outInfo->type = BLOWFISH_ENCRYPT;
		outInfo->quality = 1.0;		
		outInfo->capability = 1.0;	
		strcpy(outInfo->name, "Blowfish encryption");
		strcpy(outInfo->MIME, "application/x-blowfish");
		status = B_OK;
	}
	else if(outType == BLOWFISH_DECRYPT)
	{
		outInfo->group = B_TRANSLATOR_NONE;
		outInfo->type = BLOWFISH_DECRYPT;
		outInfo->quality = 1.0;		
		outInfo->capability = 1.0;	
		strcpy(outInfo->name, "Blowfish decryption");
		strcpy(outInfo->MIME, "application/x-blowfish");
		status = B_OK;
	}
	
	return status;
}

status_t Translate(BPositionIO * inSource,
				   const translator_info *inInfo,
				   BMessage * ioExtension,
				   uint32 outType,
				   BPositionIO * outDestination)
{	
	status_t status = B_NO_TRANSLATOR;
	
	inSource->Seek(0, SEEK_SET);	/* paranoia */
	inInfo = inInfo;				/* silence compiler warning */
	
	if(ioExtension != NULL)
	{
		const char *key = (const char *)malloc(256);
		ioExtension->FindString("key", &key);
		
		if(outType == BLOWFISH_ENCRYPT)
		{
			Encrypt(inSource, outDestination, key);	
			status = B_OK;
		}
		else if(outType == BLOWFISH_DECRYPT)
		{
			Decrypt(inSource, outDestination, key);	
			status = B_OK;
		}
	}
	
	return status;
}


void Encrypt(BPositionIO *inSource, BPositionIO *outSource, const char *key)
{
	MakeKeys(key);
	
	char buffer[BLOCK_SIZE];
	int32 bytesRead = 0;
	while((bytesRead = inSource->Read((void *)buffer, BLOCK_SIZE)) == BLOCK_SIZE)
	{	
		int32 L = B_HOST_TO_BENDIAN_INT32(*((int32 *)&buffer[0]));
		int32 R = B_HOST_TO_BENDIAN_INT32(*((int32 *)&buffer[4]));
		
		int32 a, b, c, d;
	
		L ^= P[0];
		for(int32 i = 0 ; i < rounds ; ) 
		{
			a =          (L >> 24) & 0xFF;
			b = 0x100 | ((L >> 16) & 0xFF);		// 256 +
			c = 0x200 | ((L >>  8) & 0xFF);		// 512 +
			d = 0x300 |   L         & 0xFF;		// 768 +
			R ^= (((sKey[a] + sKey[b]) ^ sKey[c]) + sKey[d]) ^ P[++i];
		
			a =          (R >> 24) & 0xFF;
			b = 0x100 | ((R >> 16) & 0xFF);		// 256 +
			c = 0x200 | ((R >>  8) & 0xFF);		// 512 +
			d = 0x300 |   R         & 0xFF;		// 768 +
			L ^= (((sKey[a] + sKey[b]) ^ sKey[c]) + sKey[d]) ^ P[++i];
		}
		R ^= P[rounds + 1];
		
		L = B_BENDIAN_TO_HOST_INT32(L);
		R = B_BENDIAN_TO_HOST_INT32(R);
		
		outSource->Write((const void *)&R, (BLOCK_SIZE / 2));
		outSource->Write((const void *)&L, (BLOCK_SIZE / 2));
	}
	
	//write any extra bytes that are left over 
	if(bytesRead > 0 && bytesRead < BLOCK_SIZE)
	{
		outSource->Write((const void *)buffer, bytesRead);	
	}
}

void Decrypt(BPositionIO *inSource, BPositionIO *outSource, const char *key)
{
	MakeKeys(key);
	
	char buffer[BLOCK_SIZE];
	int32 bytesRead = 0;
	while((bytesRead = inSource->Read((void *)buffer, BLOCK_SIZE)) == BLOCK_SIZE)
	{
		int32 L = B_HOST_TO_BENDIAN_INT32(*((int32 *)&buffer[0]));
		int32 R = B_HOST_TO_BENDIAN_INT32(*((int32 *)&buffer[4]));
		
		int a, b, c, d;
		
		L ^= P[rounds + 1];
		for (int i = rounds; i > 0; ) 
		{
			a =          (L >> 24) & 0xFF;
			b = 0x100 | ((L >> 16) & 0xFF);		// 256 +
			c = 0x200 | ((L >>  8) & 0xFF);		// 512 +
			d = 0x300 |   L        & 0xFF;		// 768 +
			R ^= (((sKey[a] + sKey[b]) ^ sKey[c]) + sKey[d]) ^ P[i--];

			a =          (R >> 24) & 0xFF;
			b = 0x100 | ((R >> 16) & 0xFF);		// 256 +
			c = 0x200 | ((R >>  8) & 0xFF);		// 512 +
			d = 0x300 |   R        & 0xFF;		// 768 +
			L ^= (((sKey[a] + sKey[b]) ^ sKey[c]) + sKey[d]) ^ P[i--];
		}
		R ^= P[0];
 		
 		L = B_BENDIAN_TO_HOST_INT32(L);
		R = B_BENDIAN_TO_HOST_INT32(R);
		
 		outSource->Write((const void *)&R, (BLOCK_SIZE / 2));
 		outSource->Write((const void *)&L, (BLOCK_SIZE / 2));
	}
	
	//write any extra bytes that are left over 
	if(bytesRead > 0 && bytesRead < BLOCK_SIZE)
	{
		outSource->Write((const void *)buffer, bytesRead);	
	} 
}

void MakeKeys(const char *key)
{
	int32 len = 0;		
	if((len = strlen(key)) > 0)
	{	
		int32 ri;
		int32 j = 0;
		for(int32 i = 0 ; i < rounds + 2 ; i++) 
		{
			ri = 0;
			for(int32 k = 0; k < 4; k++) 
			{
				ri = (ri << 8) | (key[j++] & 0xFF);
				j %= len;
			}
			P[i] = Pi[i] ^ ri;
		}
		
		//populate the session s-boxes with the inital values
		for(int32 i = 0 ; i < (4 * 256) ; i++)
		{
			sKey[i] = sInit[i];
		}
		
		BF_encrypt(0, 0, P, 0);
		
		for(int32 i = 2 ; i < rounds + 2 ; i += 2)
		{
			BF_encrypt(P[i - 2], P[i - 1], P, i);
		}
		
		//and this session s-boxes
		BF_encrypt(P[rounds], P[rounds + 1], sKey, 0);
		for(int32 i = 2 ; i < (4 * 256) ; i += 2)
		{
			BF_encrypt(sKey[i - 2], sKey[i - 1], sKey, i);
		}
	}
}

void BF_encrypt(int32 L, int32 R, int32 *out, int32 outOff) 
{
	int32 a, b, c, d;
	
	L ^= P[0];
	for(int32 i = 0 ; i < rounds ; ) 
	{
		a =          (L >> 24) & 0xFF;
		b = 0x100 | ((L >> 16) & 0xFF);		// 256 +
		c = 0x200 | ((L >>  8) & 0xFF);		// 512 +
		d = 0x300 |   L         & 0xFF;		// 768 +
		R ^= (((sKey[a] + sKey[b]) ^ sKey[c]) + sKey[d]) ^ P[++i];
		
		a =          (R >> 24) & 0xFF;
		b = 0x100 | ((R >> 16) & 0xFF);		// 256 +
		c = 0x200 | ((R >>  8) & 0xFF);		// 512 +
		d = 0x300 |   R         & 0xFF;		// 768 +
		L ^= (((sKey[a] + sKey[b]) ^ sKey[c]) + sKey[d]) ^ P[++i];
	}
	R ^= P[rounds + 1];
	
	out[outOff    ] = R;
	out[outOff + 1] = L;
}